@@ -1,7 +1,7 @@
 # NetworkManager - a library to make interacting with the NetworkManager daemon
 # easier.
 #
-# (C)2011-2013 Dennis Kaarsemaker
+# (C)2011-2015 Dennis Kaarsemaker
 # License: GPL3+
 
 import dbus
@@ -96,6 +96,8 @@ class NMDbusInterface(object):
             if hasattr(val, 'items'):
                 return dict([(x, self.wrap(y)) for x, y in val.items()])
             else:
+                if (isinstance(val, dbus.Struct) or isinstance(val, dbus.ByteArray)):
+                    return val
                 return [self.wrap(x) for x in val]
         return val
 
@@ -141,17 +143,26 @@ class NetworkManager(NMDbusInterface):
             for key in settings:
                 if 'mac-address' in settings[key]:
                     settings[key]['mac-address'] = fixups.mac_to_dbus(settings[key]['mac-address'])
+                if 'cloned-mac-address' in settings[key]:
+                    settings[key]['cloned-mac-address'] = fixups.mac_to_dbus(settings[key]['cloned-mac-address'])
                 if 'bssid' in settings[key]:
                     settings[key]['bssid'] = fixups.mac_to_dbus(settings[key]['mac-address'])
             if 'ssid' in settings.get('802-11-wireless', {}):
                 settings['802-11-wireless']['ssid'] = fixups.ssid_to_dbus(settings['802-11-wireless']['ssid'])
             if 'ipv4' in settings:
                 if 'addresses' in settings['ipv4']:
-                    settings['ipv4']['addresses'] = [fixups.addrconf_to_dbus(addr) for addr in settings['ipv4']['addresses']]
+                    settings['ipv4']['addresses'] = [fixups.addrconf_to_dbus(addr,socket.AF_INET) for addr in settings['ipv4']['addresses']]
                 if 'routes' in settings['ipv4']:
-                    settings['ipv4']['routes'] = [fixups.route_to_dbus(route) for route in settings['ipv4']['routes']]
+                    settings['ipv4']['routes'] = [fixups.route_to_dbus(route,socket.AF_INET) for route in settings['ipv4']['routes']]
                 if 'dns' in settings['ipv4']:
-                    settings['ipv4']['dns'] = [fixups.addr_to_dbus(addr) for addr in settings['ipv4']['dns']]
+                    settings['ipv4']['dns'] = [fixups.addr_to_dbus(addr,socket.AF_INET) for addr in settings['ipv4']['dns']]
+            if 'ipv6' in settings:
+                if 'addresses' in settings['ipv6']:
+                    settings['ipv6']['addresses'] = [fixups.addrconf_to_dbus(addr,socket.AF_INET6) for addr in settings['ipv6']['addresses']]
+                if 'routes' in settings['ipv6']:
+                    settings['ipv6']['routes'] = [fixups.route_to_dbus(route,socket.AF_INET6) for route in settings['ipv6']['routes']]
+                if 'dns' in settings['ipv6']:
+                    settings['ipv6']['dns'] = [fixups.addr_to_dbus(addr,socket.AF_INET6) for addr in settings['ipv6']['dns']]
         return args, kwargs
 NetworkManager = NetworkManager()
 
@@ -174,7 +185,10 @@ class Connection(NMDbusInterface):
                     break
             else:
                 return {}
-        return self.make_proxy_call('GetSecrets')(name)
+        try:
+            return self.make_proxy_call('GetSecrets')(name)
+        except:
+            return {}
 
     def postprocess(self, name, val):
         if name == 'GetSettings':
@@ -184,12 +198,18 @@ class Connection(NMDbusInterface):
                 val_ = val[key]
                 if 'mac-address' in val_:
                     val_['mac-address'] = fixups.mac_to_python(val_['mac-address'])
+                if 'cloned-mac-address' in val_:
+                    val_['cloned-mac-address'] = fixups.mac_to_python(val_['cloned-mac-address'])
                 if 'bssid' in val_:
                     val_['bssid'] = fixups.mac_to_python(val_['bssid'])
             if 'ipv4' in val:
-                val['ipv4']['addresses'] = [fixups.addrconf_to_python(addr) for addr in val['ipv4']['addresses']]
-                val['ipv4']['routes'] = [fixups.route_to_python(route) for route in val['ipv4']['routes']]
-                val['ipv4']['dns'] = [fixups.addr_to_python(addr) for addr in val['ipv4']['dns']]
+                val['ipv4']['addresses'] = [fixups.addrconf_to_python(addr,socket.AF_INET) for addr in val['ipv4']['addresses']]
+                val['ipv4']['routes'] = [fixups.route_to_python(route,socket.AF_INET) for route in val['ipv4']['routes']]
+                val['ipv4']['dns'] = [fixups.addr_to_python(addr,socket.AF_INET) for addr in val['ipv4']['dns']]
+            if 'ipv6' in val:
+                val['ipv6']['addresses'] = [fixups.addrconf_to_python(addr,socket.AF_INET6) for addr in val['ipv6']['addresses']]
+                val['ipv6']['routes'] = [fixups.route_to_python(route,socket.AF_INET6) for route in val['ipv6']['routes']]
+                val['ipv6']['dns'] = [fixups.addr_to_python(addr,socket.AF_INET6) for addr in val['ipv6']['dns']]
         return val
     preprocess = NetworkManager.preprocess
 
@@ -212,11 +232,14 @@ class Device(NMDbusInterface):
             NM_DEVICE_TYPE_VLAN: Vlan,
             NM_DEVICE_TYPE_ADSL: Adsl,
             NM_DEVICE_TYPE_BRIDGE: Bridge,
+            NM_DEVICE_TYPE_GENERIC: Generic
         }[self.DeviceType](self.object_path)
 
     def postprocess(self, name, val):
         if name == 'Ip4Address':
-            return fixups.addr_to_python(val)
+            return fixups.addr_to_python(val,socket.AF_INET)
+        if name == 'Ip6Address':
+            return fixups.addr_to_python(val,socket.AF_INET6)
         return val
 
 class AccessPoint(NMDbusInterface):
@@ -225,6 +248,8 @@ class AccessPoint(NMDbusInterface):
     def postprocess(self, name, val):
         if name == 'Ssid':
             return fixups.ssid_to_python(val)
+        elif name == 'Strength':
+            return fixups.strength_to_python(val)
         return val
 
 class Wired(NMDbusInterface):
@@ -260,6 +285,9 @@ class Vlan(NMDbusInterface):
 class Adsl(NMDbusInterface):
     interface_name = 'org.freedesktop.NetworkManager.Device.adsl'
 
+class Generic(NMDbusInterface):
+    interface_name = 'org.freedesktop.NetworkManager.Device.Generic'
+
 class NSP(NMDbusInterface):
     interface_name = 'org.freedesktop.NetworkManager.Wimax.NSP'
 
@@ -268,16 +296,25 @@ class IP4Config(NMDbusInterface):
 
     def postprocess(self, name, val):
         if name == 'Addresses':
-            return [fixups.addrconf_to_python(addr) for addr in val]
+            return [fixups.addrconf_to_python(addr,socket.AF_INET) for addr in val]
         if name == 'Routes':
-            return [fixups.route_to_python(route) for route in val]
+            return [fixups.route_to_python(route,socket.AF_INET) for route in val]
         if name in ('Nameservers', 'WinsServers'):
-            return [fixups.addr_to_python(addr) for addr in val]
+            return [fixups.addr_to_python(addr,socket.AF_INET) for addr in val]
         return val
 
 class IP6Config(NMDbusInterface):
     interface_name = 'org.freedesktop.NetworkManager.IP6Config'
 
+    def postprocess(self, name, val):
+        if name == 'Addresses':
+            return [fixups.addrconf_to_python(addr,socket.AF_INET6) for addr in val]
+        if name == 'Routes':
+            return [fixups.route_to_python(route,socket.AF_INET6) for route in val]
+        if name in ('Nameservers', 'WinsServers'):
+            return [fixups.addr_to_python(addr,socket.AF_INET6) for addr in val]
+        return val
+
 class DHCP4Config(NMDbusInterface):
     interface_name = 'org.freedesktop.NetworkManager.DHCP4Config'
 
@@ -295,9 +332,9 @@ class VPNConnection(NMDbusInterface):
 
     def preprocess(self, name, args, kwargs):
         conf = args[0]
-        conf['addresses'] = [fixups.addrconf_to_python(addr) for addr in conf['addresses']]
-        conf['routes'] = [fixups.route_to_python(route) for route in conf['routes']]
-        conf['dns'] = [fixups.addr_to_python(addr) for addr in conf['dns']]
+        conf['addresses'] = [fixups.addrconf_to_python(addr,socket.AF_INET) for addr in conf['addresses']]
+        conf['routes'] = [fixups.route_to_python(route,socket.AF_INET) for route in conf['routes']]
+        conf['dns'] = [fixups.addr_to_python(addr,socket.AF_INET) for addr in conf['dns']]
         return args, kwargs
 
 class VPNPlugin(NMDbusInterface):
@@ -326,6 +363,10 @@ class fixups(object):
             ssid = ssid.encode('utf-8')
         return [dbus.Byte(x) for x in ssid]
 
+    @staticmethod
+    def strength_to_python(strength):
+        return struct.unpack('B', strength)[0]
+
     @staticmethod
     def mac_to_python(mac):
         return "%02X:%02X:%02X:%02X:%02X:%02X" % tuple([ord(x) for x in mac])
@@ -335,52 +376,67 @@ class fixups(object):
         return [dbus.Byte(int(x, 16)) for x in mac.split(':')]
 
     @staticmethod
-    def addrconf_to_python(addrconf):
+    def addrconf_to_python(addrconf,family):
         addr, netmask, gateway = addrconf
         return [
-            fixups.addr_to_python(addr),
+            fixups.addr_to_python(addr,family),
             netmask,
-            fixups.addr_to_python(gateway)
+            fixups.addr_to_python(gateway,family)
         ]
 
     @staticmethod
-    def addrconf_to_dbus(addrconf):
+    def addrconf_to_dbus(addrconf,family):
         addr, netmask, gateway = addrconf
-        return [
-            fixups.addr_to_dbus(addr),
-            fixups.mask_to_dbus(netmask),
-            fixups.addr_to_dbus(gateway)
-        ]
+        if (family == socket.AF_INET):
+            return [
+                fixups.addr_to_dbus(addr,family),
+                fixups.mask_to_dbus(netmask),
+                fixups.addr_to_dbus(gateway,family)
+            ]
+        else:
+            return dbus.Struct(
+                (
+                    fixups.addr_to_dbus(addr,family),
+                    fixups.mask_to_dbus(netmask),
+                    fixups.addr_to_dbus(gateway,family)
+                ), signature = 'ayuay'
+            )
 
     @staticmethod
-    def addr_to_python(addr):
-        return socket.inet_ntoa(struct.pack('I', addr))
+    def addr_to_python(addr,family):
+        if (family == socket.AF_INET):
+            return socket.inet_ntop(family,struct.pack('I', addr))
+        else:
+            return socket.inet_ntop(family,b''.join(addr))
 
     @staticmethod
-    def addr_to_dbus(addr):
-        return dbus.UInt32(struct.unpack('I', socket.inet_aton(addr))[0])
+    def addr_to_dbus(addr,family):
+        if (family == socket.AF_INET):
+            return dbus.UInt32(struct.unpack('I', socket.inet_pton(family,addr))[0])
+        else:
+            return dbus.ByteArray(socket.inet_pton(family,addr))
 
     @staticmethod
     def mask_to_dbus(mask):
         return dbus.UInt32(mask)
 
     @staticmethod
-    def route_to_python(route):
+    def route_to_python(route,family):
         addr, netmask, gateway, metric = route
         return [
-            fixups.addr_to_python(addr),
+            fixups.addr_to_python(addr,family),
             netmask,
-            fixups.addr_to_python(gateway),
+            fixups.addr_to_python(gateway,family),
             socket.ntohl(metric)
         ]
 
     @staticmethod
-    def route_to_dbus(route):
+    def route_to_dbus(route,family):
         addr, netmask, gateway, metric = route
         return [
-            fixups.addr_to_dbus(addr),
+            fixups.addr_to_dbus(addr,family),
             fixups.mask_to_dbus(netmask),
-            fixups.addr_to_dbus(gateway),
+            fixups.addr_to_dbus(gateway,family),
             socket.htonl(metric)
         ]
 
@@ -412,9 +468,12 @@ NM_DEVICE_TYPE_BOND = 10
 NM_DEVICE_TYPE_VLAN = 11
 NM_DEVICE_TYPE_ADSL = 12
 NM_DEVICE_TYPE_BRIDGE = 13
+NM_DEVICE_TYPE_GENERIC = 14
+NM_DEVICE_TYPE_TEAM = 15
 NM_DEVICE_CAP_NONE = 0
 NM_DEVICE_CAP_NM_SUPPORTED = 1
 NM_DEVICE_CAP_CARRIER_DETECT = 2
+NM_DEVICE_CAP_IS_SOFTWARE = 4
 NM_WIFI_DEVICE_CAP_NONE = 0
 NM_WIFI_DEVICE_CAP_CIPHER_WEP40 = 1
 NM_WIFI_DEVICE_CAP_CIPHER_WEP104 = 2
@@ -424,7 +483,9 @@ NM_WIFI_DEVICE_CAP_WPA = 16
 NM_WIFI_DEVICE_CAP_RSN = 32
 NM_WIFI_DEVICE_CAP_AP = 64
 NM_WIFI_DEVICE_CAP_ADHOC = 128
-NM_WIFI_DEVICE_CAP_IBSS_RSN = 256
+NM_WIFI_DEVICE_CAP_FREQ_VALID = 256
+NM_WIFI_DEVICE_CAP_FREQ_2GHZ = 512
+NM_WIFI_DEVICE_CAP_FREQ_5GHZ = 1024
 NM_802_11_AP_FLAGS_NONE = 0
 NM_802_11_AP_FLAGS_PRIVACY = 1
 NM_802_11_AP_SEC_NONE = 0
@@ -450,7 +511,6 @@ NM_DEVICE_MODEM_CAPABILITY_POTS = 1
 NM_DEVICE_MODEM_CAPABILITY_CDMA_EVDO = 2
 NM_DEVICE_MODEM_CAPABILITY_GSM_UMTS = 4
 NM_DEVICE_MODEM_CAPABILITY_LTE = 8
-NM_DEVICE_MODEM_CAPABILITY_OFONO = 16
 NM_DEVICE_STATE_UNKNOWN = 0
 NM_DEVICE_STATE_UNMANAGED = 10
 NM_DEVICE_STATE_UNAVAILABLE = 20
@@ -519,6 +579,14 @@ NM_DEVICE_STATE_REASON_BR2684_FAILED = 51
 NM_DEVICE_STATE_REASON_MODEM_MANAGER_UNAVAILABLE = 52
 NM_DEVICE_STATE_REASON_SSID_NOT_FOUND = 53
 NM_DEVICE_STATE_REASON_SECONDARY_CONNECTION_FAILED = 54
+NM_DEVICE_STATE_REASON_DCB_FCOE_FAILED = 55
+NM_DEVICE_STATE_REASON_TEAMD_CONTROL_FAILED = 56
+NM_DEVICE_STATE_REASON_MODEM_FAILED = 57
+NM_DEVICE_STATE_REASON_MODEM_AVAILABLE = 58
+NM_DEVICE_STATE_REASON_SIM_PIN_INCORRECT = 59
+NM_DEVICE_STATE_REASON_NEW_ACTIVATION = 60
+NM_DEVICE_STATE_REASON_PARENT_CHANGED = 61
+NM_DEVICE_STATE_REASON_PARENT_MANAGED_CHANGED = 62
 NM_DEVICE_STATE_REASON_LAST = 65535
 NM_ACTIVE_CONNECTION_STATE_UNKNOWN = 0
 NM_ACTIVE_CONNECTION_STATE_ACTIVATING = 1
